var SmMessageEncoder = {
  REGEXP_ATTACHMENT: /<a class="attachment" href="(.*?)" .*?<\/a>|<img.*?src="((?=file:).*?)".*?>/g,

  createAttachmentHtml: function(url, name, contentType) {
    var isImage = contentType.match(/^image/);
    return "<a class=\"attachment\" href=\"" + url + "\" title=\"" + name + "\">" + 
           (isImage ? "<img src=\"" + url + "\">" : "<span>" + name + "</span>") +
           "</a>";
  },

  encode: function(message) {
    // Base64 encode header
    function encodeHeader(string) {
      if (string && message.charset) {
        for(var i in string) {
          // Do not encode string if it contains only English characters
          if (string.charCodeAt(i) > 127) {
            return "=?" + message.charset + "?B?" +
                   btoa(SmUtils.fromUTF8(string, message.charset)) + "?=";
          }
        }
      }
      return string;
    }

    function addHeader(name, value) {
      headers += name + ": " + value + "\r\n";
    }

    function addAddressHeader(name, value) {
      if (value && message.charset) {
        var addresses = SmAddress.parseList(value);
        for(var i in addresses) {
          addresses[i].name = encodeHeader(addresses[i].name);
        }
        value = addresses.join(", ");
      }
      if (value) addHeader(name, value);
    }

    // String concatenation on large (10 Mb) strings is time consuming,
    // it's better to write chunks to a temporary file and then read it back into memory
    function addCRLFs(string) {
      const CHUNK_LEN = 75;

      var tmpFile = SmFile.createTemporaryFile();
      var index = 0;
      var chunk;

      while(chunk = string.substr(index, CHUNK_LEN)) {
        tmpFile.write(chunk, chunk.length);
        tmpFile.write("\r\n", 2);
        index += CHUNK_LEN;
      }
      string = tmpFile.readFully();
      tmpFile.remove();
      return string;
    }

    function encodeBody() {
      var attachments = new Array();
      message.html = message.html.replace(SmMessageEncoder.REGEXP_ATTACHMENT,
                       function($0, $1, $2) {
                         var url = SmFile.getNsiURL($1 || $2);
                         var name = decodeURIComponent(url.fileName);
                         var mimeType = SmFile.getMimeType(url);
                         attachments.push({ url: url,
                                            name: name,
                                            mimeType: mimeType });
                         var newUrl = SmFile.getAttachmentURL(message.attachmentsDir, name);
                         return SmMessageEncoder.createAttachmentHtml(newUrl, name, mimeType);
                       });

      var html = message.html.replace(SmMessageEncoder.REGEXP_ATTACHMENT, "");

      //Replace "bare LFs" (some mail servers don't accept them
      html = html.replace(/(^|[^\r])\n+/g, function($0) {
                                             return $0.replace(/\n/g, "\r\n");
                                           });
      // Replace <br> with <br>\r\n to beautify result html
      html = html.replace(/(<br.*?>)(?!\r)/g, "$1\r\n");

      var body = "--" + boundary + "\r\n" +
                 "Content-Type: text/html; charset=" + message.charset + "\r\n" +
                 "Content-Transfer-Encoding: 8bit\r\n" +
                 "\r\n" + SmUtils.fromUTF8(html, message.charset) + "\r\n";

      for(var i in attachments) {
        var name = attachments[i].name;
        var contents = SmFile.readURL(attachments[i].url.spec);
        SmFile.saveAttachment(message.attachmentsDir, name, contents);
        name = encodeHeader(name);

        body += "--" + boundary + "\r\n" +
                "Content-Disposition: attachment; filename=\"" + name + "\"\r\n" +
                "Content-Type: " + attachments[i].mimeType +";" +
                " name=\"" + name + "\"\r\n" +
                "Content-Transfer-Encoding: base64\r\n" + "\r\n" +
                addCRLFs(btoa(contents)) + "\r\n";
      }
      return body + "--" + boundary + "--\r\n";
    }

    var headers = "";
    addAddressHeader("From", message.from);
    addAddressHeader("To", message.to);
    addAddressHeader("Cc", message.cc);
    addAddressHeader("Bcc", message.bcc);
    addHeader("Subject", encodeHeader(message.subject));
    addHeader("Date", message.date);
    addHeader("User-Agent", "Simple Mail 2.32");
    var boundary = "----------simplemail" + new Date().getTime();
    addHeader("Content-Type", "multipart/mixed; boundary=" + boundary);
    return headers + "\r\n" + encodeBody();
  },

  decode: function(source) {
    function Part(source) {
      var self = this;
      var headers = getHeaders();

      function getHeaders() {
        var start = source.indexOf("\r\n\r\n") + 2; // Leave \r\n for the last header
        return source.substr(0, start);
      }

      this.getHeader = function(name) {
        var result = new RegExp(name + ":\\s((.|\\r\\n\\s)*)\\r\\n", "i").exec(headers); // Headers may be multiline (RFC 2822: "folding")
        if (result) {
          result = result[1].replace(/\r\n(?=\s)/g, ""); // Unfold (RFC 2822)
          return decodeHeader(result);
        }
        return "";
      }

      // Removes redundant quotes, etc.
      function cleanAddressList(list) {
        var addresses = SmAddress.parseList(list);
        return addresses.join(", ");
      }

      this.getAddressHeader = function(name) {
        return cleanAddressList(self.getHeader(name));
      }

      // Decode Quoted Printable string
      function qp(string) {
        return string.replace(/=\r\n|=(..)/g, function($0, $1) {
                                          return ($0 != "=\r\n") ? String.fromCharCode("0x" + $1) : "";
                                        });
      }

      // Decode Base64 / Quoted Printable encoded header
      function decodeHeader(string) {
        if (string) {
          var regexp = /=\?(.*?)\?([BbQq])\?(.*?)\?=/g;
          string = string.replace(regexp, function($0, $1, $2, $3) {
                                            return SmUtils.toUTF8($2 == 'B' || $2 == 'b' ? atob($3) : qp($3).replace(/_/g, " "), $1);
                                          });
        }
        return string;
      }

      function getByRegexp(regexp) {
        var result = regexp.exec(headers);
        return result ? result[1] : "";
      }

      function getContentType() {
        return getByRegexp(/Content-Type: ([^;\s]*)/i);
      }

      function getBoundary() {
        return getByRegexp(/boundary=['"]?([^;\s'"]*)/i);
      }

      this.getCharset = function() {
        return getByRegexp(/charset=['"]?([^;\s'"]*)/i);
      }

      function getFileName() {
        var regexp = /(file)?name=['"]?([^;\r'"]*)/i; // File name may contain spaces
        var result = regexp.exec(headers);
        return result ? decodeHeader(result[2]) : "";
      }


      function isAttachment() {
        var regexp = /Content-Disposition: attachment/i;
        return headers.match(regexp);
      }

      // String.replace(/\r\n/g, "") on large (10 Mb) strings is time consuming,
      // it's better to split() string, write chunks to a temporary file and then read it back into memory
      function stripCRLFs(string) {
        var tmpFile = SmFile.createTemporaryFile();
        var chunks = string.split(/\r\n/);
        for(var i in chunks) {
          tmpFile.write(chunks[i], chunks[i].length);
        }
        string = tmpFile.readFully();
        tmpFile.remove();
        return string;
      }

      function getBody() {
        var start = source.indexOf("\r\n\r\n") + 4;
        var body = source.substr(start);

        var encoding = self.getHeader("Content-Transfer-Encoding");

        body = encoding.match(/base64/i) ? atob(stripCRLFs(body)) :
               encoding.match(/quoted-printable/i) ? qp(body) : body;

        var charset = self.getCharset();
        if (!message.charset) message.charset = charset;
        return charset ? SmUtils.toUTF8(body, charset) : body;
      }

      function parseMultipart() {
        var html = "";
        var parts = source.split("--" + getBoundary());

        // For multipart/alternative take the latest alternate part
        var i = getContentType().match(/multipart\/alternative/i) ? parts.length - 2 : 1;

        for(; i < parts.length - 1; i++) {
          html += " " + new Part(parts[i]).toHtml();
        }
        return html;
      }
      
      this.toHtml = function() {
        try {
          var contentType = getContentType();
          if (contentType.match(/^multipart/i)) return parseMultipart();
          else {
            var body = getBody();
            if (!isAttachment()) {
              if (!contentType || contentType.match(/text\/plain/i)) {
                return SmUtils.text2html(body); // consider empty content type as "text/plain"
              }
              else if (contentType.match(/text\/html/i)) {
                // A hacker may send a message containing <a class="attachment"> or <img>
                // tag with a link to local file, thus if such message is replied,
                // this local file would be sent as attachment in reply.
                // Filter out attachment tags to prevent this type of attack.
                return body.replace(SmMessageEncoder.REGEXP_ATTACHMENT, "");
              }
            }
            var name = getFileName();
            if (!name) name = new Date().getTime();
            SmFile.saveAttachment(message.attachmentsDir, name, body);

            var url = SmFile.getAttachmentURL(message.attachmentsDir, name);
            return SmMessageEncoder.createAttachmentHtml(url, name, contentType);
          }
        } catch(e) {
          var error = SmUtils.error("messageParseError", [e]);
          return "<font color='red'>" + error + "</font>";
        }
      }
    }
    var part = new Part(source);
    var message = new SmMessage();
    message.source = source;
    message.subject = part.getHeader("Subject");
    message.from = part.getAddressHeader("From");
    message.to = part.getAddressHeader("To");
    message.cc = part.getAddressHeader("\nCc"); // Do not match "Bcc"
    message.bcc = part.getAddressHeader("Bcc");
    message.date = part.getHeader("Date");
    message.size = source.length;
    message.html = part.toHtml();
    return message;
  }
}
